#Standard Library
import os
import re
import logging
import json

#Installed
import yaml
import requests


from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

#Default Logger
logging.basicConfig()
py_logger = logging.getLogger("pcpi")
py_logger.setLevel(10)

#Local
from .saas_session_manager import SaaSSessionManager
from .onprem_session_manager import CWPSessionManager

def __c_print(*args, **kwargs):
    '''
    Uses ascii codes to enable colored print statements. Works on Mac, Linux and Windows terminals
    '''

    #Magic that makes colors work on windows terminals
    os.system('')
    
    #Define Colors for more readable output
    c_gray = '\033[90m'
    c_red = '\033[91m'
    c_green = '\033[92m'
    c_yellow = '\033[93m'
    c_blue = '\033[94m'
    c_end = '\033[0m'

    color = c_end
    if 'color' in kwargs:
        c = kwargs['color'].lower()
        if c == 'gray' or c == 'grey':
            color = c_gray
        elif c ==  'red':
            color = c_red
        elif c == 'green':
            color = c_green
        elif c == 'yellow':
            color = c_yellow
        elif c == 'blue':
            color = c_blue
        else:
            color = c_end

    _end = '\n'
    if 'end' in kwargs:
        _end = kwargs['end']

    print(f'{color}', end='')
    for val in args:
        print(val, end='')
    print(f'{c_end}', end=_end)

#==============================================================================

def __validate_cwp_credentials(name, _url, uname, passwd, verify) -> bool:
    '''
    This function creates a session with the supplied credentials to test 
    if the user successfully entered valid credentials.
    '''

    headers = {
    'content-type': 'application/json; charset=UTF-8'
    }

    payload = {
        "username": uname,
        "password": passwd,
    }

    url = f'{_url}/api/v1/authenticate'

    try:
        __c_print('API - Validating credentials')
        res = requests.request("POST", url, headers=headers, json=payload, verify=verify)
        print(res.status_code)

        if res.status_code == 200:
            __c_print('SUCCESS', color='green')
            print()
            return True
        else:
            return False
    except:
        __c_print('ERROR', end=' ', color='red')
        print('Could not connect to Prisma Cloud API.')
        print()
        print('Steps to troubleshoot:')
        __c_print('1) Please disconnect from any incompatible VPN', color='blue')
        print()
        __c_print('2) Please ensure you have entered a valid Prisma Cloud URL.', color='blue')
        print()
        return False

def __validate_credentials(a_key, s_key, url, verify) -> bool:
    '''
    This function creates a session with the supplied credentials to test 
    if the user successfully entered valid credentials.
    '''

    headers = {
    'content-type': 'application/json; charset=UTF-8'
    }

    payload = {
        "username": f"{a_key}",
        "password": f"{s_key}"
    }

    try:
        __c_print('API - Validating credentials')
        response = requests.request("POST", f'{url}/login', headers=headers, json=payload, verify=verify)

        if response.status_code == 200:
            __c_print('SUCCESS', color='green')
            print()
            return True
        else:
            return False
    except:
        __c_print('ERROR', end=' ', color='red')
        print('Could not connect to Prisma Cloud API.')
        print()
        print('Steps to troubleshoot:')
        __c_print('1) Please disconnect from any incompatible VPN', color='blue')
        print()
        __c_print('2) Please ensure you have entered a valid Prisma Cloud URL.', color='blue')
        print('EX: https://app.prismacloud.io or https://app2.eu.prismacloud.io')
        print()
        return False

def __universal_validate_credentials(name, url, _id, secret, verify):
    if verify.lower() == 'true':
        verify = True
    else:
        verify = False
    if 'prismacloud.io' in url or 'prismacloud.cn' in url:
        return __validate_credentials(_id, secret, url, verify)
    else:
        return __validate_cwp_credentials(name, url, _id, secret, verify)

#==============================================================================

def __validate_url(url):
    if "prismacloud.io" not in (url):
        if 'https://' not in url and 'http://' not in url:
            url = 'https://' + url
        return url

    if len(url) >= 3:
        if 'https://' not in url:
            if url[:3] == 'app' or url[:3] == 'api':
                url = 'https://' + url
            
    url = url.replace('app', 'api')

    url = re.sub(r'prismacloud\.io\S*', 'prismacloud.io', url)

    return url

#==============================================================================

def __get_cwp_tenant_credentials():

    __c_print('Enter console name or any preferred identifier (optional):', color='blue')
    name = input()

    __c_print('Enter console base url with port number:', color='blue')
    url = input()
    print()

    __c_print('Enter console username:', color='blue')
    uname = input()
    print()

    __c_print('Enter console password:', color='blue')
    passwd = input()
    print()

    __c_print('Certificate verification: (True/False/<path to .pem file>)', color='blue')
    __c_print('Leave blank to use default value.', color='yellow')
    verify = input()
    print()

    if verify == '':
        verify = True
    elif verify.lower() == 'true':
        verify = True
    elif verify.lower() == 'false':
        verify = False
    else:
        pass

    return name, url, uname, passwd, verify

def __get_config():
    __c_print('Enter Prisma URL. (SaaS EX: https://app.ca.prismacloud.io, On-Prem EX: https://yourdomain.com):', color='blue')
    url = input()
    print()
    new_url = __validate_url(url)
    if new_url != url:
        __c_print('Adjusted URL:',color='yellow')
        print(new_url)
        print()

    __c_print('Enter identity (Access Key or Username):', color='blue')
    _id = input()
    print()

    __c_print('Enter secret (Secret Key or Password):', color='blue')
    secret = input()
    print()

    __c_print('Certificate verification: (True/False/<path_to_.pem> file)', color='blue')
    __c_print('Leave blank to use default value (True).', color='yellow')
    verify = input()
    print()

    #If there is non-prisma URL, then ask if its a self hosted project
    project_flag = 'false'
    if 'prismacloud.io' not in new_url and 'prismacloud.cn' not in new_url:
        __c_print('CWP Project (True/False)', color='blue')
        __c_print('Leave blank to use default value (False).', color='yellow')
        project_flag = input()
        print()
        if project_flag.lower() == 'true':
            project_flag = 'true'
        else:
            project_flag = 'false'
    
    name = 'DEFAULT_NAME'
    if project_flag == 'true':
        __c_print('Enter project ID:', color='blue')
        name = input()
    else:
        __c_print('Enter tenant/console name (Optional):', color='blue')
        name = input()

    print()

    verify = verify.strip()

    if verify == '':
        verify = 'true'
    elif verify.lower() == 'true':
        verify = 'true'
    elif verify.lower() == 'false':
        verify = 'false'
    else:
        pass
    

    return name, _id, secret, new_url, verify, project_flag

def __get_tenant_credentials():

    __c_print('Enter tenant name or any preferred identifier (optional):', color='blue')
    name = input()

    __c_print('Enter tenant url. (ex: https://app.ca.prismacloud.io):', color='blue')
    url = input()
    print()
    new_url = __validate_url(url)
    if new_url != url:
        __c_print('Adjusted URL:',color='yellow')
        print(new_url)
        print()

    __c_print('Enter tenant access key:', color='blue')
    a_key = input()
    print()

    __c_print('Enter tenant secret key:', color='blue')
    s_key = input()
    print()

    __c_print('Certificate verification: (True/False/<path to .pem file>)', color='blue')
    __c_print('Leave blank to use default value.', color='yellow')
    verify = input()
    print()

    if verify == '':
        verify = True
    elif verify.lower() == 'true':
        verify = True
    elif verify.lower() == 'false':
        verify = False
    else:
        pass
    

    return name, a_key, s_key, new_url, verify

#==============================================================================

def __build_cwp_session_dict(name, url, uname, passwd, verify):
    session_dict = {
        name: {
            'url': url,
            'uname': uname,
            'passwd': passwd,
            'verify': verify
            }
    }
    return session_dict

def __build_session_dict(name, a_key, s_key, url, verify):
    session_dict = {
        name: {
            'access_key': a_key,
            'secret_key': s_key,
            'api_url': url,
            'verify': verify
            }
    }
    return session_dict

def __build_config_json(name, _id, secret, url, verify, project_flag):
    session_dict = {
        'name': name,
        'url': url,
        'identity': _id,
        'secret': secret,
        'verify': verify,
        'project_flag': project_flag
    }
    return session_dict

#==============================================================================

def __get_min_cwp_credentials_from_user(min_tenants):
    credentials = []
    tenants_added = 0

    while True:
        valid = False
        while not valid:
            __c_print('Enter credentials for the console', color='blue')
            print()
            name, url, uname, passwd, verify = __get_cwp_tenant_credentials()
            
            valid = __validate_cwp_credentials(name, url, uname, passwd, verify)
            if valid == False:
                __c_print('FAILED', end=' ', color='red')
                print('Invalid credentials. Please re-enter your credentials')
                print()
            else:
                credentials.append(__build_cwp_session_dict(name, url, uname, passwd, verify))
                tenants_added += 1
        
        if tenants_added >= min_tenants:
            __c_print('Would you like to add an other tenant? Y/N')
            choice = input().lower()

            if choice != 'yes' and choice != 'y':
                break

    return credentials

def __get_cwp_credentials_from_user(num_tenants):
    #Gets the source tenant credentials and ensures that are valid
    credentials = []

    if num_tenants != -1:
        for i in range(num_tenants):
            valid = False
            while not valid:
                __c_print('Enter credentials for the console', color='blue')
                print()
                name, url, uname, passwd, verify = __get_cwp_tenant_credentials()
                
                valid = __validate_cwp_credentials(name, url, uname, passwd, verify)
                if valid == False:
                    __c_print('FAILED', end=' ', color='red')
                    print('Invalid credentials. Please re-enter your credentials')
                    print()
                else:
                    credentials.append(__build_cwp_session_dict(name, url, uname, passwd, verify))

        return credentials
    else:
        while True:
            valid = False
            while not valid:
                __c_print('Enter credentials for the console', color='blue')
                print()
                name, url, uname, passwd, verify = __get_cwp_tenant_credentials()
                
                valid = __validate_cwp_credentials(name, url, uname, passwd, verify)
                if valid == False:
                    __c_print('FAILED', end=' ', color='red')
                    print('Invalid credentials. Please re-enter your credentials')
                    print()
                else:
                    credentials.append(__build_cwp_session_dict(name, url, uname, passwd, verify))
            
            __c_print('Would you like to add an other tenant? Y/N')
            choice = input().lower()

            if choice != 'yes' and choice != 'y':
                break

        return credentials

def __get_min_credentials_from_user(min_tenants):
    credentials = []
    tenants_added = 0
    while True:
        valid = False
        while not valid:
            __c_print('Enter credentials for the tenant', color='blue')
            print()
            src_name, src_a_key, src_s_key, src_url, verify = __get_tenant_credentials()
            
            valid = __validate_credentials(src_a_key, src_s_key, src_url, verify)
            if valid == False:
                __c_print('FAILED', end=' ', color='red')
                print('Invalid credentials. Please re-enter your credentials')
                print()
            else:
                credentials.append(__build_session_dict(src_name, src_a_key, src_s_key, src_url, verify))
                tenants_added += 1
        
        if tenants_added >= min_tenants:
            __c_print('Would you like to add an other tenant? Y/N')
            choice = input().lower()

            if choice != 'yes' and choice != 'y':
                break

    return credentials

def __get_credentials_from_user(num_tenants):
    #Gets the source tenant credentials and ensures that are valid
    credentials = []

    if num_tenants != -1:
        for i in range(num_tenants):
            valid = False
            while not valid:
                __c_print('Enter credentials for the tenant', color='blue')
                print()
                src_name, src_a_key, src_s_key, src_url, verify = __get_tenant_credentials()
                
                valid = __validate_credentials(src_a_key, src_s_key, src_url, verify)
                if valid == False:
                    __c_print('FAILED', end=' ', color='red')
                    print('Invalid credentials. Please re-enter your credentials')
                    print()
                else:
                    credentials.append(__build_session_dict(src_name, src_a_key, src_s_key, src_url, verify))

        return credentials
    else:
        while True:
            valid = False
            while not valid:
                __c_print('Enter credentials for the tenant', color='blue')
                print()
                src_name, src_a_key, src_s_key, src_url, verify = __get_tenant_credentials()
                
                valid = __validate_credentials(src_a_key, src_s_key, src_url, verify)
                if valid == False:
                    __c_print('FAILED', end=' ', color='red')
                    print('Invalid credentials. Please re-enter your credentials')
                    print()
                else:
                    credentials.append(__build_session_dict(src_name, src_a_key, src_s_key, src_url, verify))
            
            __c_print('Would you like to add an other tenant? Y/N')
            choice = input().lower()

            if choice != 'yes' and choice != 'y':
                break

        return credentials

def __get_config_from_user(num_tenants, min_tenants):
        #Gets the source tenant credentials and ensures that are valid
    credentials = []

    if num_tenants != -1 and min_tenants == -1:
        for i in range(num_tenants):
            valid = False
            while not valid:
                __c_print('Enter Prisma Cloud Credentials', color='blue')
                print()
                name, _id, secret, url, verify, project_flag = __get_config()
                
                valid = __universal_validate_credentials(name, url, _id, secret, verify)
                if valid == False:
                    __c_print('FAILED', end=' ', color='red')
                    print('Invalid credentials. Please re-enter your credentials')
                    print()
                else:
                    credentials.append(__build_config_json(name, _id, secret, url, verify, project_flag))
    elif num_tenants == -1 and min_tenants != -1:
        tenant_count = 0
        while True:
            valid = False
            while not valid:
                __c_print('Enter Prisma Cloud Credentials', color='blue')
                print()
                name, _id, secret, url, verify, project_flag = __get_config()
                
                valid = __universal_validate_credentials(name, url, _id, secret, verify)
                if valid == False:
                    __c_print('FAILED', end=' ', color='red')
                    print('Invalid credentials. Please re-enter your credentials')
                    print()
                else:
                    credentials.append(__build_config_json(name, _id, secret, url, verify, project_flag))
                    tenant_count +=1
            
            if tenant_count >= min_tenants:
                __c_print('Would you like to add another Prisma Cloud credential? Y/N')
                choice = input().lower()
                if choice != 'yes' and choice != 'y':
                    break
    else:
        while True:
            valid = False
            while not valid:
                __c_print('Enter Prisma Cloud Credentials', color='blue')
                print()
                name, _id, secret, url, verify, project_flag = __get_config()
                
                valid = __universal_validate_credentials(name, url, _id, secret, verify)
                if valid == False:
                    __c_print('FAILED', end=' ', color='red')
                    print('Invalid credentials. Please re-enter your credentials')
                    print()
                else:
                    credentials.append(__build_config_json(name, _id, secret, url, verify, project_flag))
            
            __c_print('Would you like to add another Prisma Cloud credential? Y/N')
            choice = input().lower()

            if choice != 'yes' and choice != 'y':
                break

    return credentials


def __load_uuid_yaml(file_name, logger=py_logger):
    with open(file_name, "r") as file:
        cfg = yaml.load(file, Loader=yaml.BaseLoader)

    credentials = cfg['credentials']
    entity_type = cfg['type']
    uuid = cfg['uuid']
    cmp_type = cfg['cmp_type']

    tenant_sessions = []
    for tenant in credentials:
        tenant_name = ''
        tenant_keys = tenant.keys()
        for name in tenant_keys:
            tenant_name = name     

        a_key = tenant[tenant_name]['access_key']
        s_key = tenant[tenant_name]['secret_key']
        api_url = tenant[tenant_name]['api_url']

        tenant_sessions.append(SaaSSessionManager(tenant_name, a_key, s_key, api_url, logger))

    return tenant_sessions, entity_type, uuid, cmp_type

#==============================================================================

def onprem_load_from_env(logger=py_logger) -> object:
    error_exit = False

    name = 'Console'
    try:
        name = os.environ['PC_CONSOLE_NAME']
    except:
        logger.warning('Missing \'PC_CONSOLE_NAME\' environment variable. Using default name.')
    
    api_url = ''
    api = None
    try:
        api_url = os.environ['PC_CONSOLE_URL']
        api = __validate_url(api_url)
    except:
        logger.error('Missing \'PC_CONSOLE_URL\' environment variable.')
        error_exit = True

    uname = None
    try:
        uname = os.environ['PC_CONSOLE_USERNAME']
    except:
        logger.error('Missing \'PC_CONSOLE_USERNAME\' environment variable.')
        error_exit = True

    passwd = None
    try:
        passwd = os.environ['PC_CONSOLE_PASSWORD']
    except:
        logger.error('Missing \'PC_CONSOLE_PASSWORD\' environment variable.')
        error_exit = True

    verify = True
    try:
        verify = os.environ['PC_API_VERIFY']
        if verify.lower() == 'false':
            verify = False
        if verify.lower() == 'true':
            verify = True
    except:
        logger.warning('Missing \'PC_API_VERIFY\' environment variable. Using default value...')

    if error_exit:
        logger.info('Missing required environment variables. Exiting...')
        exit()

    return CWPSessionManager(name, api_url, uname, passwd, verify, False, logger)

#==============================================================================
def onprem_load_min_from_file(min_tenants, file_path='console_credentials.yml', logger=py_logger):
    '''
    Reads console_credentials.yml or specified file path to load
    self hosted CWP console credentials to create a session.
    Returns a CWP session object.
    '''
    #Open and load config file
    if not os.path.exists(file_path):
        #Create credentials yml file
        __c_print('No credentials file found. Generating...', color='yellow')
        print()
        tenants = __get_min_cwp_credentials_from_user(min_tenants)
        with open(file_path, 'w') as yml_file: 
            for tenant in tenants:
                yaml.dump(tenant, yml_file, default_flow_style=False)

    cfg = {}
    with open(file_path, "r") as file:
        cfg = yaml.load(file, Loader=yaml.BaseLoader)

    #Parse cfg for tenant names and create tokens for each tenant
    tenant_sessions = []
    for tenant in cfg:
        uname = cfg[tenant]['uname']
        passwd = cfg[tenant]['passwd']
        api_url = cfg[tenant]['url']
        verify = True
        try:
            verify = cfg[tenant]['verify']
            if verify.lower() == 'false':
                verify = False
            if verify.lower() == 'true':
                verify = True
        except:
            pass

        tenant_sessions.append(CWPSessionManager(tenant, api_url, uname=uname, passwd=passwd, verify=verify, project_flag=False, logger=logger))

    return tenant_sessions

def onprem_load_multi_from_file(file_path='console_credentials.yml', logger=py_logger, num_tenants=-1) -> list:
    '''
    Reads console_credentials.yml or specified file path to load
    self hosted CWP console credentials to create a session.
    Returns a CWP session object.
    '''
    #Open and load config file
    if not os.path.exists(file_path):
        #Create credentials yml file
        __c_print('No credentials file found. Generating...', color='yellow')
        print()
        tenants = __get_cwp_credentials_from_user(num_tenants)
        with open(file_path, 'w') as yml_file: 
            for tenant in tenants:
                yaml.dump(tenant, yml_file, default_flow_style=False)

    cfg = {}
    with open(file_path, "r") as file:
        cfg = yaml.load(file, Loader=yaml.BaseLoader)

    #Parse cfg for tenant names and create tokens for each tenant
    tenant_sessions = []
    for tenant in cfg:
        uname = cfg[tenant]['uname']
        passwd = cfg[tenant]['passwd']
        api_url = cfg[tenant]['url']
        verify = True
        try:
            verify = cfg[tenant]['verify']
            if verify.lower() == 'false':
                verify = False
            if verify.lower() == 'true':
                verify = True
        except:
            pass

        tenant_sessions.append(CWPSessionManager(tenant, api_url, uname=uname, passwd=passwd, verify=verify, project_flag=False, logger=logger))


    return tenant_sessions

def onprem_load_from_file(file_path='console_credentials.yml', logger=py_logger) -> list:
    '''
    Reads console_credentials.yml or specified file path to load
    self hosted CWP console credentials to create a session.
    Returns a CWP session object.
    '''
    #Open and load config file
    if not os.path.exists(file_path):
        #Create credentials yml file
        __c_print('No credentials file found. Generating...', color='yellow')
        print()
        tenants = __get_cwp_credentials_from_user(1)
        with open(file_path, 'w') as yml_file: 
            for tenant in tenants:
                yaml.dump(tenant, yml_file, default_flow_style=False)

    cfg = {}
    with open(file_path, "r") as file:
        cfg = yaml.load(file, Loader=yaml.BaseLoader)

    #Parse cfg for tenant names and create tokens for each tenant
    tenant_sessions = []
    for tenant in cfg:
        uname = cfg[tenant]['uname']
        passwd = cfg[tenant]['passwd']
        api_url = cfg[tenant]['url']
        verify = True
        try:
            verify = cfg[tenant]['verify']
            if verify.lower() == 'false':
                verify = False
            if verify.lower() == 'true':
                verify = True
        except:
            pass

        tenant_sessions.append(CWPSessionManager(tenant, api_url, uname=uname, passwd=passwd, verify=verify, project_flag=False, logger=logger))

    try:   
        return tenant_sessions[0]
    except:
        logger.error('Error - No credentials found. Exiting...')
        exit()

def load_min_from_file(min_tenants, file_path='tenant_credentials.yml', logger=py_logger) -> list:
    '''
    Reads config.yml and generates a Session object for the tenant
    Returns:
    Tenant Session object
    '''
    #Open and load config file
    if not os.path.exists(file_path):
        #Create credentials yml file
        __c_print('No credentials file found. Generating...', color='yellow')
        print()
        tenants = __get_min_credentials_from_user(min_tenants)
        with open(file_path, 'w') as yml_file:
            for tenant in tenants:
                yaml.dump(tenant, yml_file, default_flow_style=False)

    with open(file_path, "r") as file:
        cfg = yaml.load(file, Loader=yaml.BaseLoader)

    #Parse cfg for tenant names and create tokens for each tenant
    tenant_sessions = []
    for tenant in cfg:
        a_key = cfg[tenant]['access_key']
        s_key = cfg[tenant]['secret_key']
        api_url = cfg[tenant]['api_url']
        verify = True
        try:
            verify = cfg[tenant]['verify']
            if verify.lower() == 'false':
                verify = False
            if verify.lower() == 'true':
                verify = True
        except:
            pass

        tenant_sessions.append(SaaSSessionManager(tenant, a_key, s_key, api_url, verify, logger))
       
    return tenant_sessions

def load_multi_from_file(file_path='tenant_credentials.yml', logger=py_logger, num_tenants=-1) -> list:
    '''
    Reads config.yml and generates a Session object for the tenant
    Returns:
    Tenant Session object
    '''
    #Open and load config file
    if not os.path.exists(file_path):
        #Create credentials yml file
        __c_print('No credentials file found. Generating...', color='yellow')
        print()
        tenants = __get_credentials_from_user(num_tenants)
        with open(file_path, 'w') as yml_file:
            for tenant in tenants:
                yaml.dump(tenant, yml_file, default_flow_style=False)

    with open(file_path, "r") as file:
        cfg = yaml.load(file, Loader=yaml.BaseLoader)

    #Parse cfg for tenant names and create tokens for each tenant
    tenant_sessions = []
    for tenant in cfg:
        a_key = cfg[tenant]['access_key']
        s_key = cfg[tenant]['secret_key']
        api_url = cfg[tenant]['api_url']
        verify = True
        try:
            verify = cfg[tenant]['verify']
            if verify.lower() == 'false':
                verify = False
            if verify.lower() == 'true':
                verify = True
        except:
            pass

        tenant_sessions.append(SaaSSessionManager(tenant, a_key, s_key, api_url, verify, logger))
       

    return tenant_sessions

def load_from_file(file_path='tenant_credentials.yml', logger=py_logger) -> list:
    '''
    Reads config.yml and generates a Session object for the tenant
    Returns:
    Tenant Session object
    '''
    #Open and load config file
    if not os.path.exists(file_path):
        #Create credentials yml file
        __c_print('No credentials file found. Generating...', color='yellow')
        print()
        tenants = __get_credentials_from_user(1)
        with open(file_path, 'w') as yml_file: 
            for tenant in tenants:
                yaml.dump(tenant, yml_file, default_flow_style=False)

    with open(file_path, "r") as file:
        cfg = yaml.load(file, Loader=yaml.BaseLoader)

    #Parse cfg for tenant names and create tokens for each tenant
    tenant_sessions = []
    for tenant in cfg:
        a_key = cfg[tenant]['access_key']
        s_key = cfg[tenant]['secret_key']
        api_url = cfg[tenant]['api_url']
        verify = True
        try:
            verify = cfg[tenant]['verify']
            if verify.lower() == 'false':
                verify = False
            if verify.lower() == 'true':
                verify = True
        except:
            pass

        tenant_sessions.append(SaaSSessionManager(tenant, a_key, s_key, api_url, verify, logger))

    try:   
        return tenant_sessions[0]
    except:
        logger.error('Error - No credentials found. Exiting...')
        exit()

def load_from_env(logger=py_logger) -> object:
    error_exit = False

    name = 'Tenant'
    try:
        name = os.environ['PC_TENANT_NAME']
    except:
        logger.warning('Missing \'PC_TENANT_NAME\' environment variable. Using default name.')
    
    api_url = ''
    api = None
    try:
        api_url = os.environ['PC_TENANT_API']
        api = __validate_url(api_url)
    except:
        logger.error('Missing \'PC_TENANT_API\' environment variable.')
        error_exit = True

    a_key = None
    try:
        a_key = os.environ['PC_TENANT_A_KEY']
    except:
        logger.error('Missing \'PC_TENANT_A_KEY\' environment variable.')
        error_exit = True

    s_key = None
    try:
        s_key = os.environ['PC_TENANT_S_KEY']
    except:
        logger.error('Missing \'PC_TENANT_S_KEY\' environment variable.')
        error_exit = True

    verify = True
    try:
        verify = os.environ['PC_API_VERIFY']
        if verify.lower() == 'false':
            verify = False
        if verify.lower() == 'true':
            verify = True
    except:
        logger.warning('Missing \'PC_API_VERIFY\' environment variable. Using default value...')

    if error_exit:
        logger.info('Missing required environment variables. Exiting...')
        exit()

    return SaaSSessionManager(name, a_key, s_key, api_url, verify, logger)

def load_min_from_user(min_tenants, logger=py_logger):
    tenant_sessions = []
    tenants = __get_min_credentials_from_user(min_tenants)
    for tenant in tenants:
        for key in tenant:
            name = key
            verify = True
            try:
                verify = tenant[name]['verify']
                if verify.lower() == 'false':
                    verify = False
                if verify.lower() == 'true':
                    verify = True
            except:
                pass

            tenant_sessions.append(SaaSSessionManager(name, tenant[name]['access_key'], tenant[name]['secret_key'], tenant[name]['api_url'], verify, logger))
            
    return tenant_sessions

def load_multi_from_user(logger=py_logger, num_tenants=-1) -> list:
    tenant_sessions = []
    tenants = __get_credentials_from_user(num_tenants)
    for tenant in tenants:
        for key in tenant:
            name = key
            verify = True
            try:
                verify = tenant[name]['verify']
                if verify.lower() == 'false':
                    verify = False
                if verify.lower() == 'true':
                    verify = True
            except:
                pass

            tenant_sessions.append(SaaSSessionManager(name, tenant[name]['access_key'], tenant[name]['secret_key'], tenant[name]['api_url'], verify, logger))
            
    return tenant_sessions

def load_from_user(logger=py_logger) -> list:
    tenant_sessions = []
    tenants = __get_credentials_from_user(1)
    for tenant in tenants:
        for key in tenant:
            name = key
            verify = True
            try:
                verify = tenant[name]['verify']
                if verify.lower() == 'false':
                    verify = False
                if verify.lower() == 'true':
                    verify = True
            except:
                pass

            tenant_sessions.append(SaaSSessionManager(name, tenant[name]['access_key'], tenant[name]['secret_key'], tenant[name]['api_url'], verify, logger))
            
    return tenant_sessions[0]

def load_config(file_path='', num_tenants=-1, min_tenants=-1, logger=py_logger):
    if num_tenants != -1 and min_tenants != -1:
        logger.error('ERROR: Incompatible options. Exiting...')
        exit()

    if file_path == '':
        config_dir = os.path.join(os.environ['HOME'], '.prismacloud')
        config_path = os.path.join(config_dir, 'credentials.json')
        if not os.path.exists(config_dir):
            os.mkdir(config_dir)
    else:
        config_path = file_path

    tenant_sessions = []

    if not os.path.exists(config_path):
        config = __get_config_from_user(num_tenants, min_tenants)
        with open(config_path, 'w') as outfile:
            json.dump(config, outfile)

    config_data = {}
    with open(config_path, 'r') as infile:
        try:
            config_data = json.load(infile)
        except:
            logger.error('Failed to load credentials file. Exiting...')
            exit()

    for blob in config_data:
        verify = blob['verify']
        if verify.lower().strip() == 'false':
            verify = False
        elif verify.lower().strip() == 'true':
            verify = True
        else:
            verify = verify

        if 'prismacloud.io' in blob['url'] or 'prismacloud.cn' in blob['url']:
            tenant_sessions.append(SaaSSessionManager(blob['name'], blob['identity'], blob['secret'], blob['url'], verify, logger=logger))
        else:
            project_flag = False
            project_flag_in = blob.get('project_flag', None)
            if project_flag_in:
                if project_flag_in.lower().strip() == 'true':
                    project_flag = True

            tenant_sessions.append(CWPSessionManager(blob['name'], blob['url'], blob['identity'], blob['secret'], verify, project_flag, logger=logger))

    return tenant_sessions

def load_config_user(num_tenants=-1, min_tenants=-1, logger=py_logger):
    if num_tenants != -1 and min_tenants != -1:
        logger.error('ERROR: Incompatible options. Exiting...')
        exit()

    config = __get_config_from_user(num_tenants, min_tenants)

    tenant_sessions = []

    for tenant in config:
        if 'prismacloud.io' in tenant['url'] or 'prismacloud.cn' in tenant['url']:
            tenant_sessions.append(SaaSSessionManager(tenant['name'], tenant['identity'], tenant['secret'], tenant['url'], tenant['verify'], logger=logger))
        else:
            project_flag = False
            project_flag_in = tenant.get('project_flag', None)
            if project_flag_in:
                if project_flag_in.lower().strip() == 'true':
                    project_flag = True
            tenant_sessions.append(CWPSessionManager(tenant['name'], tenant['url'], tenant['identity'], tenant['secret'], tenant['verify'], project_flag, logger=logger))

    return tenant_sessions

def load_config_env(prisma_name='PRISMA_PCPI_NAME', identifier_name='PRISMA_PCPI_ID', secret_name='PRISMA_PCPI_SECRET', api_url_name='PRISMA_PCPI_URL', verify_name='PRISMA_PCPI_VERIFY', project_flag_name='PRISMA_PCPI_PROJECT_FLAG',  logger=py_logger):
    error_exit = False

    name = 'Tenant'
    try:
        name = os.environ[prisma_name]
    except:
        logger.warning(f'Missing \'{prisma_name}\' environment variable. Using default name.')
    
    api_url = ''
    api = None
    try:
        api_url = os.environ[api_url_name]
        api = __validate_url(api_url)
    except:
        logger.error(f'Missing \'{api_url_name}\' environment variable.')
        error_exit = True

    a_key = None
    try:
        a_key = os.environ[identifier_name]
    except:
        logger.error(f'Missing \'{identifier_name}\' environment variable.')
        error_exit = True

    s_key = None
    try:
        s_key = os.environ[secret_name]
    except:
        logger.error(f'Missing \'{secret_name}\' environment variable.')
        error_exit = True

    verify = True
    try:
        verify = os.environ[verify_name]
        if verify.lower() == 'false':
            verify = False
        if verify.lower() == 'true':
            verify = True
    except:
        logger.warning(f'Missing \'{verify_name}\' environment variable. Using default value...')
    
    project_flag =  False
    try:
        project_flag = os.environ[project_flag_name]
        if project_flag.lower().strip() == 'true':
            project_flag = True
    except:
        logger.warning(f'Missing \'{project_flag_name}\' environment variable. Using default value...')

    if error_exit:
        logger.info('Missing required environment variables. Exiting...')
        exit()

    if 'prismacloud.io' in api or 'prismacloud.cn' in api:
        return SaaSSessionManager(name, a_key, s_key, api, verify, logger=logger)
    else:
        return CWPSessionManager(name, api, a_key, s_key, verify, project_flag, logger=logger)



#Uncomment a section to test

###INSTALLED FROM PYPI EXAMPLES
#Minimum setup-----------------------------------------------------------------
# from pcpi import session_loader

# session_managers = session_loader.load_config()
# session_manager = session_managers[0]
# cspm_session = session_manager.create_cspm_session()

# res = cspm_session.request('GET', '/cloud')
# print(res.json())

#Recommended setup-------------------------------------------------------------
# from pcpi import session_loader
# import loguru

# session_managers = session_loader.load_config(logger=loguru.logger)
# session_manager = session_managers[0]
# cspm_session = session_manager.create_cspm_session()

# res = cspm_session.request('GET', '/cloud')
# print(res.json())

#CSPM and CWP------------------------------------------------------------------
# from pcpi import session_loader
# import loguru

# session_managers = session_loader.load_config(logger=loguru.logger)
# session_manager = session_managers[0]

# cspm_session = session_manager.create_cspm_session()
# cwp_session = session_manager.create_cwp_session()

# res = cspm_session.request('GET', '/cloud')
# print(res.json())

# print('--------------')

# res2 = cwp_session.request('GET', 'api/v1/credentials')
# print(res2.json())

#Session self healing----------------------------------------------------------
# from pcpi import session_loader
# import loguru

# session_managers = session_loader.load_config(logger=loguru.logger)
# session_manager = session_managers[0]

# cwp_session = session_manager.create_cwp_session()
# cwp_session.cspm_token = 'asdasdasd'
# cwp_session.headers['Authorization'] = 'Bearer ' + 'sdfsdfsdfsdf'

# res = cwp_session.request('GET', 'api/v1/credentials')
# print(res.json())

#Error and debugging output----------------------------------------------------------
# from pcpi import session_loader
# import loguru

# session_managers = session_loader.load_config(logger=loguru.logger)
# session_manager = session_managers[0]

# cwp_session = session_manager.create_cwp_session()
# cwp_session.cspm_token = 'asdasdasd'
# cwp_session.headers['Authorization'] = 'Bearer ' + 'sdfsdfsdfsdf'

# res = cwp_session.request('GET', 'api/v9/credentials')


###LOCAL CLONED REPO EXAMPLES==================================================

#Minimum setup-----------------------------------------------------------------
# from src.pcpi import session_loader

# session_managers = session_loader.load_config()
# session_manager = session_managers[0]
# cspm_session = session_manager.create_cspm_session()

# res = cspm_session.request('GET', '/cloud')
# print(res.json())

#Recommended setup-------------------------------------------------------------
# from src.pcpi import session_loader
# import loguru

# session_managers = session_loader.load_config(logger=loguru.logger)
# session_manager = session_managers[0]
# cspm_session = session_manager.create_cspm_session()

# res = cspm_session.request('GET', '/cloud')
# print(res.json())

#CSPM and CWP------------------------------------------------------------------
# from src.pcpi import session_loader
# import loguru

# session_managers = session_loader.load_config(logger=loguru.logger)
# session_manager = session_managers[0]

# cspm_session = session_manager.create_cspm_session()
# cwp_session = session_manager.create_cwp_session()

# res = cspm_session.request('GET', '/cloud')
# print(res.json())

# print('--------------')

# res2 = cwp_session.request('GET', 'api/v1/credentials')
# print(res2.json())

#Session self healing----------------------------------------------------------
# from src.pcpi import session_loader
# import loguru

# session_managers = session_loader.load_config(logger=loguru.logger)
# session_manager = session_managers[0]

# cwp_session = session_manager.create_cwp_session()
# cwp_session.cspm_token = 'asdasdasd'
# cwp_session.headers['Authorization'] = 'Bearer ' + 'sdfsdfsdfsdf'

# res = cwp_session.request('GET', 'api/v1/credentials')
# print(res.json())

# #Error and debugging output----------------------------------------------------------
# from src.pcpi import session_loader
# import loguru

# session_managers = session_loader.load_config(logger=loguru.logger)
# session_manager = session_managers[0]

# cwp_session = session_manager.create_cwp_session()
# cwp_session.cspm_token = 'asdasdasd'
# cwp_session.headers['Authorization'] = 'Bearer ' + 'sdfsdfsdfsdf'

# res = cwp_session.request('GET', 'api/v9/credentials')


#Heuristic Search--------------------------------------------------------------
# #Change RQL query as needed
# query = "config from cloud.resource where resource.status = Active AND api.name = 'aws-ec2-describe-instances'"

# #Payload can use absolute time range or relative time range. Examples are given of both time ranges.
# payload = {
#     "query": query,
#     "timeRange": {
#         "relativeTimeType": "BACKWARD",
#         "type": "relative",
#         "value": {
#             "amount": 24,
#             "unit": "hour"
#         }
#     },
# }


# # # Option 1 PCPI paginates automatically and returns JSON data as one chunk.
# # json_data = cspm_session.config_search_request(payload)

# # #Dump json to file
# # with open('totalDataRETURNED.json', 'w') as outfile:
# #     json.dump(json_data, outfile)

# #---------------------------

# # Option 2 PCPI paginates automatically but now you pass in a custom function that controls what to do with each page of data.
# def dump_data(details, data, counter, total_rows):
#     if counter == 0:
#         with open(f'detailsOut.json', 'w') as outfile:
#             json.dump(details, outfile)

#     with open(f'temp/dataOut{counter}.json', 'w') as outfile:
#         json.dump(data, outfile)

# total_rows = cspm_session.config_search_request_function(payload, dump_data)

# print(f'Got {total_rows} Total Rows')



import unittest
from unittest import mock
from unittest import TestCase
import os
import yaml

#Default Logger
import logging
logging.basicConfig()
py_logger = logging.getLogger("pcpi")
py_logger.setLevel(10)

#HELPER FUNCTIONS==============================================================
def load_environment():
    cfg = {}
    with open('tenant_credentials.yml', "r") as file:
        cfg = yaml.load(file, Loader=yaml.BaseLoader)

    #Parse cfg for tenant creds and set env
    for index,tenant in enumerate(cfg):
        uname = cfg[tenant]['access_key']
        passwd = cfg[tenant]['secret_key']
        api_url = cfg[tenant]['api_url']
        verify = True
        try:
            verify = cfg[tenant]['verify']
            if verify.lower() == 'false':
                verify = False
            if verify.lower() == 'true':
                verify = True
        except:
            pass
        
        os.environ[f'PC_TENANT_NAME{index}'] = tenant
        os.environ[f'PC_TENANT_API{index}'] = api_url
        os.environ[f'PC_TENANT_A_KEY{index}'] = uname
        os.environ[f'PC_TENANT_S_KEY{index}'] = passwd
        os.environ[f'PC_TENANT_VERIFY{index}'] = str(verify)

def load_onprem_environment():
    cfg = {}
    with open('console_credentials.yml', "r") as file:
        cfg = yaml.load(file, Loader=yaml.BaseLoader)

    #Parse cfg for tenant names and create tokens for each tenant
    tenant_sessions = []
    for index,tenant in enumerate(cfg):
        uname = cfg[tenant]['uname']
        passwd = cfg[tenant]['passwd']
        api_url = cfg[tenant]['url']
        verify = True
        try:
            verify = cfg[tenant]['verify']
            if verify.lower() == 'false':
                verify = False
            if verify.lower() == 'true':
                verify = True
        except:
            pass
    
        os.environ[f'PC_CONSOLE_NAME{index}'] = tenant
        os.environ[f'PC_CONSOLE_API{index}'] = api_url
        os.environ[f'PC_CONSOLE_UNAME{index}'] = uname
        os.environ[f'PC_CONSOLE_PASSWD{index}'] = passwd
        os.environ[f'PC_CONSOLE_VERIFY{index}'] = str(verify)
    
#UNIT TESTS====================================================================

class credentialFileTests(TestCase):
    def testLoadConfigBasic(self):
        load_environment()
        from src.pcpi import session_loader
        from src.pcpi import saas_session_manager
        name = os.environ['PC_TENANT_NAME0']
        api_url = os.environ['PC_TENANT_API0']
        a_key = os.environ['PC_TENANT_A_KEY0']
        s_key = os.environ['PC_TENANT_S_KEY0']
        verify = os.environ['PC_TENANT_VERIFY0']

        result = session_loader.load_config()
        self.assertEqual([result[0].tenant], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).tenant])
        self.assertEqual([result[0].a_key], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).a_key])
        self.assertEqual([result[0].s_key], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).s_key])
        self.assertEqual([result[0].api_url], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).api_url])

    def testLoadConfigFilePath(self):
        load_environment()
        from src.pcpi import session_loader
        from src.pcpi import saas_session_manager
        name = os.environ['PC_TENANT_NAME0']
        api_url = os.environ['PC_TENANT_API0']
        a_key = os.environ['PC_TENANT_A_KEY0']
        s_key = os.environ['PC_TENANT_S_KEY0']
        verify = os.environ['PC_TENANT_VERIFY0']

        result = session_loader.load_config(file_path='creds.json')
        self.assertEqual([result[0].tenant], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).tenant])
        self.assertEqual([result[0].a_key], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).a_key])
        self.assertEqual([result[0].s_key], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).s_key])
        self.assertEqual([result[0].api_url], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).api_url])

    def testLoadConfigEnv(self):
        load_environment()
        from src.pcpi import session_loader
        from src.pcpi import saas_session_manager
        name = os.environ['PC_TENANT_NAME0']
        api_url = os.environ['PC_TENANT_API0']
        a_key = os.environ['PC_TENANT_A_KEY0']
        s_key = os.environ['PC_TENANT_S_KEY0']
        verify = os.environ['PC_TENANT_VERIFY0']

        result = session_loader.load_config_env(prisma_name='PC_TENANT_NAME0', identifier_name='PC_TENANT_A_KEY0', secret_name='PC_TENANT_S_KEY0', api_url_name='PC_TENANT_API0', verify_name='PC_TENANT_VERIFY0')
        self.assertEqual(result.tenant, saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).tenant)
        self.assertEqual(result.a_key, saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).a_key)
        self.assertEqual(result.s_key, saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).s_key)
        self.assertEqual(result.api_url, saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).api_url)

    def testLoadMinFromFile(self):
        load_environment()
        from src.pcpi import session_loader
        from src.pcpi import saas_session_manager
        name = os.environ['PC_TENANT_NAME0']
        api_url = os.environ['PC_TENANT_API0']
        a_key = os.environ['PC_TENANT_A_KEY0']
        s_key = os.environ['PC_TENANT_S_KEY0']
        verify = os.environ['PC_TENANT_VERIFY0']

        name2 = os.environ['PC_TENANT_NAME1']
        api_url2 = os.environ['PC_TENANT_API1']
        a_key2 = os.environ['PC_TENANT_A_KEY1']
        s_key2 = os.environ['PC_TENANT_S_KEY1']
        verify2 = os.environ['PC_TENANT_VERIFY1']

        result = session_loader.load_min_from_file(2)
        self.assertEqual([result[0].tenant, result[1].tenant], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).tenant, saas_session_manager.SaaSSessionManager(name2, a_key2, s_key2, api_url2, verify2, py_logger).tenant])
        self.assertEqual([result[0].a_key, result[1].a_key], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).a_key, saas_session_manager.SaaSSessionManager(name2, a_key2, s_key2, api_url2, verify2, py_logger).a_key])
        self.assertEqual([result[0].s_key, result[1].s_key], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).s_key, saas_session_manager.SaaSSessionManager(name2, a_key2, s_key2, api_url2, verify2, py_logger).s_key])
        self.assertEqual([result[0].api_url, result[1].api_url], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).api_url, saas_session_manager.SaaSSessionManager(name2, a_key2, s_key2, api_url2, verify2, py_logger).api_url])

    def testLoadMultiFromFile(self):
        load_environment()
        from src.pcpi import session_loader
        from src.pcpi import saas_session_manager
        name = os.environ['PC_TENANT_NAME0']
        api_url = os.environ['PC_TENANT_API0']
        a_key = os.environ['PC_TENANT_A_KEY0']
        s_key = os.environ['PC_TENANT_S_KEY0']
        verify = os.environ['PC_TENANT_VERIFY0']

        name2 = os.environ['PC_TENANT_NAME1']
        api_url2 = os.environ['PC_TENANT_API1']
        a_key2 = os.environ['PC_TENANT_A_KEY1']
        s_key2 = os.environ['PC_TENANT_S_KEY1']
        verify2 = os.environ['PC_TENANT_VERIFY1']

        result = session_loader.load_multi_from_file(num_tenants=2)
        self.assertEqual([result[0].tenant, result[1].tenant], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).tenant, saas_session_manager.SaaSSessionManager(name2, a_key2, s_key2, api_url2, verify2, py_logger).tenant])
        self.assertEqual([result[0].a_key, result[1].a_key], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).a_key, saas_session_manager.SaaSSessionManager(name2, a_key2, s_key2, api_url2, verify2, py_logger).a_key])
        self.assertEqual([result[0].s_key, result[1].s_key], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).s_key, saas_session_manager.SaaSSessionManager(name2, a_key2, s_key2, api_url2, verify2, py_logger).s_key])
        self.assertEqual([result[0].api_url, result[1].api_url], [saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).api_url, saas_session_manager.SaaSSessionManager(name2, a_key2, s_key2, api_url2, verify2, py_logger).api_url])

    def testLoadFromFile(self):
        load_environment()
        from src.pcpi import session_loader
        from src.pcpi import saas_session_manager
        name = os.environ['PC_TENANT_NAME0']
        api_url = os.environ['PC_TENANT_API0']
        a_key = os.environ['PC_TENANT_A_KEY0']
        s_key = os.environ['PC_TENANT_S_KEY0']
        verify = os.environ['PC_TENANT_VERIFY0']

        result = session_loader.load_from_file()
        self.assertEqual(result.tenant, saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).tenant)
        self.assertEqual(result.a_key, saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).a_key)
        self.assertEqual(result.s_key, saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).s_key)
        self.assertEqual(result.api_url, saas_session_manager.SaaSSessionManager(name, a_key, s_key, api_url, verify, py_logger).api_url)

    def testOnpremLoadMinFromFile(self):
        load_onprem_environment()
        from src.pcpi import session_loader
        from src.pcpi import onprem_session_manager
        name = os.environ['PC_CONSOLE_NAME0']
        api_url = os.environ['PC_CONSOLE_API0']
        uname = os.environ['PC_CONSOLE_UNAME0']
        passwd = os.environ['PC_CONSOLE_PASSWD0']
        verify =  os.environ['PC_CONSOLE_VERIFY0']

        name2 = os.environ['PC_CONSOLE_NAME1']
        api_url2 = os.environ['PC_CONSOLE_API1']
        uname2 = os.environ['PC_CONSOLE_UNAME1']
        passwd2 = os.environ['PC_CONSOLE_PASSWD1']
        verify2 =  os.environ['PC_CONSOLE_VERIFY1']

        result = session_loader.onprem_load_min_from_file(2)

        self.assertEqual([result[0].tenant, result[1].tenant], [onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).tenant, onprem_session_manager.CWPSessionManager(name2, api_url2, uname2, passwd2, verify2, False, py_logger).tenant])
        self.assertEqual([result[0].uname, result[1].uname], [onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).uname, onprem_session_manager.CWPSessionManager(name2, api_url2, uname2, passwd2, verify2, False, py_logger).uname])
        self.assertEqual([result[0].passwd, result[1].passwd], [onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).passwd, onprem_session_manager.CWPSessionManager(name2, api_url2, uname2, passwd2, verify2, False, py_logger).passwd])
        self.assertEqual([result[0].api_url, result[1].api_url], [onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).api_url, onprem_session_manager.CWPSessionManager(name2, api_url2, uname2, passwd2, verify2, False, py_logger).api_url])

    def testOnpremLoadMultiFromFile(self):
        load_onprem_environment()
        from src.pcpi import session_loader
        from src.pcpi import onprem_session_manager
        name = os.environ['PC_CONSOLE_NAME0']
        api_url = os.environ['PC_CONSOLE_API0']
        uname = os.environ['PC_CONSOLE_UNAME0']
        passwd = os.environ['PC_CONSOLE_PASSWD0']
        verify =  os.environ['PC_CONSOLE_VERIFY0']

        name2 = os.environ['PC_CONSOLE_NAME1']
        api_url2 = os.environ['PC_CONSOLE_API1']
        uname2 = os.environ['PC_CONSOLE_UNAME1']
        passwd2 = os.environ['PC_CONSOLE_PASSWD1']
        verify2 =  os.environ['PC_CONSOLE_VERIFY1']

        result = session_loader.onprem_load_multi_from_file(num_tenants=2)

        self.assertEqual([result[0].tenant, result[1].tenant], [onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).tenant, onprem_session_manager.CWPSessionManager(name2, api_url2, uname2, passwd2, verify2, False, py_logger).tenant])
        self.assertEqual([result[0].uname, result[1].uname], [onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).uname, onprem_session_manager.CWPSessionManager(name2, api_url2, uname2, passwd2, verify2, False, py_logger).uname])
        self.assertEqual([result[0].passwd, result[1].passwd], [onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).passwd, onprem_session_manager.CWPSessionManager(name2, api_url2, uname2, passwd2, verify2, False, py_logger).passwd])
        self.assertEqual([result[0].api_url, result[1].api_url], [onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).api_url, onprem_session_manager.CWPSessionManager(name2, api_url2, uname2, passwd2, verify2, False, py_logger).api_url])

    def testOnpremLoadFromFile(self):
        load_onprem_environment()
        from src.pcpi import session_loader
        from src.pcpi import onprem_session_manager
        name = os.environ['PC_CONSOLE_NAME0']
        api_url = os.environ['PC_CONSOLE_API0']
        uname = os.environ['PC_CONSOLE_UNAME0']
        passwd = os.environ['PC_CONSOLE_PASSWD0']
        verify =  os.environ['PC_CONSOLE_VERIFY0']

        result = session_loader.onprem_load_from_file()

        self.assertEqual(result.tenant, onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).tenant)
        self.assertEqual(result.uname, onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).uname)
        self.assertEqual(result.passwd, onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).passwd)
        self.assertEqual(result.api_url, onprem_session_manager.CWPSessionManager(name, api_url, uname, passwd, verify, False, py_logger).api_url)

class apiRequestTest(TestCase):
    def testCSPMRecovery(self):
        from src.pcpi import session_loader
        manager = session_loader.load_from_file()
        cspm_session = manager.create_cspm_session()
        res = cspm_session.request('GET', '/cloud')
        cspm_session.headers = {
            'content-type': 'application/json; charset=UTF-8',
            'x-redlock-auth': 'asd'
        }
        res1 = cspm_session.request('GET', '/cloud')

        self.assertEqual(res.json(), res1.json())

    def testCWPRecovery(self):
        from src.pcpi import session_loader
        manager = session_loader.load_from_file()
        cwp_session = manager.create_cwp_session()
        res = cwp_session.request('GET', '/api/v1/users')
        cwp_session.headers = {
            'content-type': 'application/json; charset=UTF-8',
            'Authorization': 'Bearer dfsdfsd'
        }
        cwp_session.cspm_token = 'sdfsdf'
        res1 = cwp_session.request('GET', '/api/v1/users')

        self.assertEqual(res.json(), res1.json())

    def testCertBypass(self):
        from src.pcpi import session_loader
        manager = session_loader.load_from_file()
        cspm_session = manager.create_cspm_session()
        res = cspm_session.request('GET', '/cloud', verify=False)

        self.assertEqual(res.status_code, 200)

    # def testCertOverwrite(self):
    #     from src.pcpi import session_loader
    #     manager = session_loader.load_from_file()
    #     cspm_session = manager.create_cspm_session()
    #     res = cspm_session.request('GET', '/compliance', verify='globalprotect_certifi.txt')

    #     self.assertEqual(res.status_code, 200)
        
if __name__ == '__main__':
    unittest.main()


#Standard Library
import enum

#Installed
import requests

from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

#Local
from ._session_base import Session

import time

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class CWPSession(Session):
    def __init__(self, tenant_name: str, api_url: str, uname: str, passwd:str, verify:bool, project_flag:bool, logger: object):
        """
        Initializes a Prisma Cloud API session for a given tenant.

        Keyword Arguments:
        tenant_name -- Name of tenant associated with session
        api_url -- API URL Tenant is hosted on
        uname -- Username
        passwd -- Password
        token -- Dedicated API token
        """

        super().__init__(logger)

        self.tenant = tenant_name
        self.uname = uname
        self.passwd = passwd
        self.verify = verify
        self.project_flag = project_flag

        self.token_time_stamp = 0

        self.api_url = api_url

        self.logger = logger

        self.auth_key = 'Authorization'
        self.auth_style = 'Bearer '

        self.headers = {
            'content-type': 'application/json; charset=UTF-8',
            'Authorization': 'Bearer '
        }

        self._api_login_wrapper()

#==============================================================================
    def _api_login(self) -> object:
        '''
        Calls the Prisma Cloud API to generate a x-redlock-auth JWT.

        Returns:
        x-redlock-auth JWT.
        '''

        #Build request
        url = f'{self.api_url}/api/v1/authenticate'
        
        headers = {
            'content-type': 'application/json; charset=UTF-8'
            }

        payload = {
            "username": self.uname,
            "password": self.passwd,
        }

        self.logger.debug('API - Generating CWP session token.')

        res = object()
        try:
            start_time = time.time()
            self.logger.debug(url)
            res = requests.request("POST", url, headers=headers, json=payload, verify=self.verify)
            end_time = time.time()
            time_completed = round(end_time-start_time, 3)

            self.token_time_stamp = time.time()
        except:
            self.logger.error('Failed to connect to API.')
            self.logger.warning('Make sure any offending VPNs are disabled.')

        return [res, time_completed]

    def _expired_login(self) -> None:
        self.logger.error('FAILED')
        self.logger.warning('Invalid Login Credentials. JWT not generated. Exiting...')
        exit()


#Installed
import requests

from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

#Local
from ._session_base import Session
from ._cspm_session import CSPMSession

#Python Library
import time

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SaaSCWPSession(Session):
    def __init__(self, tenant_name: str, a_key: str, s_key: str, api_url: str, verify, logger, cspm_session={}):
        """
        Initializes a Prisma Cloud API session for a given tenant.

        Keyword Arguments:
        tenant_name -- Name of tenant associated with session
        a_key -- Tenant Access Key
        s_key -- Tenant Secret Key
        api_url -- API URL Tenant is hosted on
        """

        super().__init__(logger)

        self.tenant = tenant_name
        self.a_key = a_key
        self.s_key = s_key
        self.api_url = api_url
        self.verify = verify

        self.token_time_stamp = 0

        self.logger = logger

        self.cspm_session = {}
        self.cspm_token = ''

        if not cspm_session:
            self.__get_cspm_session()
        else:
            self.cspm_session = cspm_session
            self.cspm_token = self.cspm_session.token

        self.api_url = self.__cwpp_metadata(self.cspm_session)

        self.auth_key = 'Authorization'
        self.auth_style = 'Bearer '
        
        self.token = ''

        self.headers = {
            'content-type': 'application/json; charset=UTF-8',
            'Authorization': 'Bearer '
        }

        self._api_login_wrapper()

#==============================================================================
    def __get_cspm_session(self):
        self.cspm_session = CSPMSession(self.tenant, self.a_key, self.s_key, self.api_url, self.verify, self.logger)
        self.cspm_token = self.cspm_session.token
    
    def __cspm_login(self):
        self.cspm_token = self.cspm_session._api_login_wrapper()

    def __cwpp_metadata(self, cspm_session):
        res = cspm_session.request('GET', 'meta_info')
        compute_url = res.json()['twistlockUrl']

        return compute_url

#==============================================================================
    def _api_login(self) -> object:
        '''
        Calls the Prisma Cloud API to generate a x-redlock-auth JWT.

        Returns:
        x-redlock-auth JWT.
        '''

        #Build request
        url = f'{self.api_url}/api/v1/authenticate'
        
        headers = {
            'content-type': 'application/json; charset=UTF-8'
            }

        payload = {
            "username": None,
            "password": None,
            "token": self.cspm_token
        }

        self.logger.debug('API - Generating SaaS CWPP session token.')

        res = object()
        try:
            start_time = time.time()
            self.logger.debug(url)
            res = requests.request("POST", url, headers=headers, json=payload, verify=self.verify)
            end_time = time.time()
            time_completed = round(end_time-start_time, 3)
            
            self.token_time_stamp = time.time()
        except:
            self.logger.error('Failed to connect to API.')
            self.logger.warning('Make sure any offending VPNs are disabled.')

        return [res, time_completed]

    def _expired_login(self) -> None:
        self.logger.warning('CSPM session expired. Generating new session.')
        self.__cspm_login()

    def _api_refresh(self) -> None:
        self.logger.debug('API - Refreshing SaaS session token.')

        res = object()
        try:
            res, time_completed = self.cspm_session._api_refresh()
            self.token_time_stamp = time.time()
        except:
            self.logger.error('Failed to connect to API.')
            self.logger.warning('Make sure any offending VPNs are disabled.')

        return [res, time_completed]


""" Utility to create a CA bundle including GlobalProtect certificates """
#Install pythons openssl library
#pip3 install pyopenssl

import argparse
import socket
import OpenSSL

import certifi

# --Description-- #

# Prisma Cloud SSL Helper (requires 'pip install certifi pyopenssl')

# --Configuration-- #

pc_arg_parser = argparse.ArgumentParser()
pc_arg_parser.add_argument(
        '--api',
        default='api.prismacloud.io',
        type=str,
        help='(Optional) - Prisma Cloud API/UI Base URL')
pc_arg_parser.add_argument(
        '--api_port',
        default=443,
        type=int,
        help='(Optional) - Prisma Cloud API/UI Port.')
pc_arg_parser.add_argument(
        '--ca_bundle',
        default='globalprotect_certifi.txt',
        type=str,
        help='(Optional) - Custom CA (bundle) file to create')
args = pc_arg_parser.parse_args()

src_ca_file = certifi.where()
dst_ca_file = args.ca_bundle
host_name = args.api
port = args.api_port
panw_subjects = [
    '/C=US/ST=CA/O=paloalto networks/OU=IT/CN=decrypt.paloaltonetworks.com',
    '/DC=local/DC=paloaltonetworks/CN=Palo Alto Networks Inc Domain CA',
    '/C=US/O=Palo Alto Networks Inc/CN=Palo Alto Networks Inc Root CA'
]
ssl_context_method = OpenSSL.SSL.TLSv1_2_METHOD # SSL.SSLv23_METHOD

# --Main-- #

with open(src_ca_file, 'r') as root_ca_file:
    root_certificates = root_ca_file.read()

with open(dst_ca_file, 'w') as custom_ca_file:
    custom_ca_file.write(root_certificates)
    context = OpenSSL.SSL.Context(method=ssl_context_method)
    context.load_verify_locations(cafile=src_ca_file)
    conn = OpenSSL.SSL.Connection(context, socket=socket.socket(socket.AF_INET, socket.SOCK_STREAM))
    conn.settimeout(5)
    conn.connect((host_name, port))
    conn.setblocking(1)
    conn.do_handshake()
    conn.set_tlsext_host_name(host_name.encode())
    for (idx, certificate) in enumerate(conn.get_peer_cert_chain()):
        subject = ''.join("/{0:s}={1:s}".format(name.decode(), value.decode()) for name, value in certificate.get_subject().get_components())
        issuer  = ''.join("/{0:s}={1:s}".format(name.decode(), value.decode()) for name, value in certificate.get_issuer().get_components())
        if subject in panw_subjects:
            subject_string     = '# Subject: %s' % subject
            issuer_string      = '# Issuer: %s' % issuer
            certificate_string = OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, certificate).decode('utf-8')
            custom_ca_file.write("\n")
            custom_ca_file.write(subject_string)
            custom_ca_file.write("\n")
            custom_ca_file.write(issuer_string)
            custom_ca_file.write("\n")
            custom_ca_file.write(certificate_string)
    conn.close()

print('CA bundle file saved as: %s' % dst_ca_file)
print("Use it with these scripts by specifying '--verify %s' on the command line or in your config file" % dst_ca_file)


#Standard Library
import time

import json as json_lib

#installed
import requests

from urllib3.exceptions import InsecureRequestWarning

class Session:
    def __init__(self,logger):
        """
        Initializes a Prisma Cloud API Session Manager.

        Keyword Arguments:
        logger - optional logger, either pylib logger or loguru
        """
        self.retries = 20
        self.retry_statuses = [401, 425, 429, 500, 502, 503, 504]
        self.retry_delay_statuses = [429, 500, 502, 503, 504]
        self.success_status = [200,201,202,203,204,205,206]
        self.expired_code = 401
        self.retry_timer = 0
        self.retry_timer_max = 32
        self.token_time = 360

        self.empty_res = ''

        self.u_count = 1
        self.unknown_error_max = 5

        self.logger = logger

#==============================================================================
    def _api_login_wrapper(self):
        if self.verify == False:
            requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

        res = self.empty_res
        while res == self.empty_res and self.u_count < self.unknown_error_max:
            try:
                res, time_completed = self._api_login()

                retries = 0
                encountered_401 = False
                while res.status_code in self.retry_statuses and retries < self.retries:
                    if res.status_code in self.retry_delay_statuses:
                        self.logger.warning(f'FAILED {res.status_code} - {time_completed} seconds')
                        
                        #Increase timer when ever encounted to slow script execution.
                        if self.retry_timer == 0:
                            self.retry_timer = 1
                        else:
                            self.retry_timer = self.retry_timer*2
                            if self.retry_timer >= self.retry_timer_max:
                                self.retry_timer = self.retry_timer_max

                        self.logger.warning(f'Waiting {self.retry_timer} seconds')
                        time.sleep(self.retry_timer)
                        self.logger.warning('Increasing wait time.')

                    elif res.status_code == self.expired_code:
                        if encountered_401 == True:
                            self.logger.error('ERROR - Can not perform API operations')
                            self.logger.warning('Steps to troubleshoot: ')
                            self.logger.warning('1) Ensure Access Key has proper permission.')
                            self.logger.warning('2) Ensure Prisma Cloud license is valid.')
                            self.logger.warning('3) If using a CWPP Project, ensure project name is valid in credential configurations.')
                            exit()

                        self._expired_login()
                        encountered_401 = True
                    else:
                        code = "NO_CODE"
                        try:
                            code = res.status_code
                        except:
                            pass
                        self.logger.error(f'FAILED {code} - {time_completed} seconds')
                        self.logger.error('ERROR Logging In. JWT not generated.')
                        self.logger.warning('RESPONSE:')
                        self.logger.info(res)
                        self.logger.warning('RESPONSE URL:')
                        self.logger.info(res.url)
                        self.logger.warning('RESPONSE TEXT:')
                        self.logger.info(res.text)

                    
                    res, time_completed = self._api_login()

                    retries +=1
                
                if retries == self.retries:
                    self.logger.error('ERROR. Max retires exceeded on API Login. Exiting...')
                    exit()

                #Update token and headers
                token = res.json().get('token')
                self.token = token
                new_headers = self.headers
                new_headers[self.auth_key] = self.auth_style + token
                self.headers = new_headers

                try:
                    self.logger.success(f'SUCCESS - {time_completed} seconds')
                except:
                    self.logger.info(f'SUCCESS - {time_completed} seconds')

                self.u_count = 1
                return token

            except KeyboardInterrupt:
                self.logger.error('Keyboard Interrupt. Exiting...')
                exit()
            except Exception as e:
                self.logger.error(e)
                self.logger.error(f'UNKNOWN ERROR - API login. Retrying... {self.u_count} of {self.unknown_error_max}')
                self.u_count += 1
                self.logger.warning('Steps to troubleshoot: ')
                self.logger.warning('1) Disable any VPNs.')
                self.logger.warning('2) Ensure API base URL is correct.')
                time.sleep(1)

    def _api_refresh_wrapper(self):
        res = self.empty_res

        refresh_func = getattr(self, "_api_refresh", None)
        if not callable(refresh_func):
            return self._api_login_wrapper()
        else:
            while res == self.empty_res and self.u_count < self.unknown_error_max:
                try:
                    res, time_completed = self._api_refresh()

                    retries = 0
                    while res.status_code in self.retry_statuses and retries < self.retries:
                        if res.status_code in self.retry_delay_statuses:
                            self.logger.warning(f'CODE {res.status_code} - {time_completed} seconds')
                            self.logger.warning(f'HEADERS: {res.headers}')
                            #Increase timer when ever encounter to slow script execution.
                            if self.retry_timer == 0:
                                self.retry_timer = 1
                            else:
                                self.retry_timer = self.retry_timer*2
                                if self.retry_timer >= self.retry_timer_max:
                                    self.retry_timer = self.retry_timer_max

                            self.logger.warning(f'Waiting {self.retry_timer} seconds')
                            time.sleep(self.retry_timer)
                            self.logger.warning('Increasing wait time.')

                        elif res.status_code == self.expired_code:
                            self.logger.error(f'FAILED {self.expired_code} - {time_completed} seconds')
                            self.logger.warning('Session expired. Generating new Token')
                            self._api_login_wrapper()
                            return
                        else:
                            code = "NO_CODE"
                            try:
                                code = res.status_code
                            except:
                                pass
                            self.logger.error(f'FAILED {code} - {time_completed} seconds')
                            self.logger.error('ERROR Refreshing Token.')
                            self.logger.warning('RESPONSE:')
                            self.logger.info(res)
                            self.logger.warning('RESPONSE URL:')
                            self.logger.info(res.url)
                            self.logger.warning('RESPONSE TEXT:')
                            self.logger.info(res.text)
                        
                        res, time_completed = self._api_refresh()

                        retries +=1
                    
                    if retries == self.retries:
                        self.logger.error('ERROR. Max retires exceeded on JWT refresh. Exiting...')
                        exit()

                    #Update token and headers
                    token = res.json().get('token')
                    self.token = token
                    new_headers = self.headers
                    new_headers[self.auth_key] = self.auth_style + token
                    self.headers = new_headers

                    try:
                        self.logger.success(f'SUCCESS - {time_completed} seconds')
                    except:
                        self.logger.info(f'SUCCESS - {time_completed} seconds')

                    self.u_count = 1

                    return

                except KeyboardInterrupt:
                    self.logger.error('Keyboard Interrupt. Exiting...')
                    exit()
                except Exception as e:
                    self.logger.error(e)
                    self.logger.error(f'UNKNOWN ERROR - API refresh. Retrying... {self.u_count} of {self.unknown_error_max}')
                    self.u_count += 1
                    self.logger.warning('Steps to troubleshoot: ')
                    self.logger.warning('1) Disable any VPNs.')
                    self.logger.warning('2) Ensure API base URL is correct.')
                    time.sleep(1)


#==============================================================================
    def __api_call_wrapper(self, method: str, url: str, json: dict=None, data: dict=None, params: dict=None, verify=True, acceptCsv=False, redlock_ignore: list=None, status_ignore: list=[], custom_log='', custom_error_message=''):
        """
        A wrapper around all API calls that handles token generation, retrying
        requests and API error console output logging.
        Keyword Arguments:
        method -- Request method/type. Ex: POST or GET
        url -- Full API request URL
        data -- Body of the request in a json compatible format
        params -- Queries for the API request
        Returns:
        Respose from API call.
        """
        res = self.empty_res
        while res == self.empty_res and self.u_count < self.unknown_error_max:
            try:
                if time.time() - self.token_time_stamp >= self.token_time:
                    self.logger.warning('Session Refresh Timer - Generating new Token')
                    self._api_refresh_wrapper()

                self.logger.debug(f'{url}')
                res,time_completed = self.__request_wrapper(method, url, headers=self.headers, json=json, data=data, params=params, verify=verify, acceptCsv=acceptCsv)
                
                if res.status_code in self.success_status or res.status_code in status_ignore:
                    log_message = ''
                    if custom_log:
                        log_message = ' - ' + custom_log

                    try:
                        self.logger.success(f'SUCCESS - {time_completed} seconds{log_message}')
                    except:
                        self.logger.info(f'SUCCESS - {time_completed} seconds{log_message}')
                    return res

                retries = 0
                encountered_401 = False
                while res.status_code in self.retry_statuses and retries < self.retries:
                    #If we get a 429 code, sleep for a doubling amount of time.
                    if res.status_code in self.retry_delay_statuses:
                        error_msg = ''
                        log_message = ''
                        if custom_error_message:
                            error_msg = " - " + custom_error_message
                        if custom_log:
                            log_message = ' - ' + custom_log

                        self.logger.warning(f'FAILED {res.status_code} - {time_completed} seconds{log_message}{error_msg}')
                        #Wait for retry timer
                        if self.retry_timer > 0:
                            self.logger.warning(f'Waiting {self.retry_timer} seconds')
                            time.sleep(self.retry_timer)

                        self.logger.warning('Increasing wait time')
                        #Increase timer when ever wait status encounted to slow script execution.
                        if self.retry_timer == 0:
                            self.retry_timer = 1
                        else:
                            self.retry_timer = self.retry_timer*2
                            if self.retry_timer >= self.retry_timer_max:
                                self.retry_timer = self.retry_timer_max
                    
                    #If token expires, login again and get new token
                    if res.status_code == self.expired_code:
                        if encountered_401 == True:
                            self.logger.error('ERROR - Can not perform API operations')
                            self.logger.warning('Steps to troubleshoot: ')
                            self.logger.warning('1) Ensure Access Key has proper permission.')
                            self.logger.warning('2) Ensure Prisma Cloud license is valid.')
                            self.logger.warning('3) If using a CWPP Project, ensure project name is valid in credential configurations.')
                            exit()

                        self.logger.error(f'FAILED {self.expired_code} - {time_completed} seconds')
                        self.logger.warning('Session expired. Generating new Token and retrying')
                        self._api_login_wrapper()
                        encountered_401 = True

                    self.logger.warning(f'Retrying request')
                    self.logger.debug(f'{url}')

                    res, time_completed = self.__request_wrapper(method, url, headers=self.headers, json=json, data=data, params=params, verify=verify, acceptCsv=acceptCsv)
                    retries += 1
                
                if res.status_code in self.success_status or res.status_code in status_ignore:
                    try:
                        self.logger.success(f'SUCCESS - {time_completed} seconds')
                    except:
                        self.logger.info(f'SUCCESS - {time_completed} seconds')
                    return res

                if retries >= self.retries:
                    self.logger.error('ERROR. Max retires exceeded')

                #Some redlock errors need to be handled elsewhere and don't require this debugging output
                if 'x-redlock-status' in res.headers and redlock_ignore:
                    for el in redlock_ignore:
                        if el in res.headers['x-redlock-status']:
                            return res

                code = "NO_CODE"
                try:
                    code = res.status_code
                except:
                    pass
                self.logger.error(f'FAILED {code} - {time_completed} seconds')
                self.logger.error('REQUEST DUMP:')
                self.logger.warning('REQUEST HEADERS:')
                self.logger.info(self.headers)
                self.logger.warning('REQUEST JSON:')
                self.logger.info(json)
                if data:
                    self.logger.warning('REQUEST DATA:')
                    self.logger.info(data)
                self.logger.warning('REQUEST PARAMS:')
                self.logger.info(params)
                self.logger.warning('RESPONSE:')
                self.logger.info(res)
                self.logger.warning('RESPONSE URL:')
                self.logger.info(res.url)
                self.logger.warning('RESPONSE HEADERS:')
                self.logger.info(res.headers)
                self.logger.warning('RESPONSE REQUEST BODY:')
                self.logger.info(res.request.body)
                self.logger.warning('RESPONSE STATUS:')
                if 'x-redlock-status' in res.headers:
                    self.logger.info(res.headers['x-redlock-status'])
                self.logger.warning('RESPONSE TEXT:')
                self.logger.info(res.text)
                self.logger.warning('RESPONSE JSON:')
                if res.text != "":
                    for json_data in res.json():
                        self.logger.info(json_data)

                self.u_count = 1
                return res
            except KeyboardInterrupt:
                self.logger.error('Keyboard Interrupt. Exiting...')
                exit()
            except Exception as e:
                #self.logger.error(e, f'- {time_completed} seconds')
                self.logger.error(e)
                if res == self.empty_res:
                    self.logger.error(f'UNKNOWN ERROR - API Call Wrapper. Retrying... {self.u_count} of {self.unknown_error_max}')
                    time.sleep(2)
                else:
                    self.logger.error(f'UNKNOWN ERROR - API Call Wrapper. Continuing... {self.u_count} of {self.unknown_error_max}')
                self.u_count += 1

    #==============================================================================

    def request(self, method: str, endpoint_url: str, json: dict=None, data: dict=None, params: dict=None, verify=None, acceptCsv=False, redlock_ignore: list=None, status_ignore: list=[], custom_log='', custom_error_message=''):
        '''
        Function for calling the PC API using this session manager. Accepts the
        same arguments as 'requests.request' minus the headers argument as 
        headers are supplied by the session manager.
        '''
        if verify == None:
            verify = self.verify

        #If the CWP Session is a project, auto add project id query string
        try:
            if self.project_flag == True:
                if params:
                    params.update({"project":self.tenant})
                else:
                    params = {"project":self.tenant}
        except:
            pass

        #Validate method
        method = method.upper()
        if method not in ['POST', 'PUT', 'GET', 'OPTIONS', 'DELETE', 'PATCH']:
            self.logger.warning('Invalid method.')
        
        #Build url
        if endpoint_url[0] != '/':
            endpoint_url = '/' + endpoint_url

        url = f'{self.api_url}{endpoint_url}'

        #Call wrapper
        return self.__api_call_wrapper(method, url, json=json, data=data, params=params, verify=verify, acceptCsv=acceptCsv, redlock_ignore=redlock_ignore, status_ignore=status_ignore, custom_log=custom_log, custom_error_message=custom_error_message)

    def config_search_request(self, json: dict, verify=None, acceptCsv=False, redlock_ignore: list=None, status_ignore: list=[]):
        if verify == None:
            verify = self.verify
        
        limit = 20000#Max limit value is 100,000
        
        #Force best practices with HS
        json.update({"heuristicSearch": True, "limit": limit, "withResourceJson": True})

        #initial API Call
        res = self.__api_call_wrapper('POST', f'{self.api_url}/search/config', json=json, verify=verify, acceptCsv=acceptCsv, redlock_ignore=redlock_ignore, status_ignore=status_ignore)

        total_rows = 0
        complete_res_list = []
        complete_res_dict = res.json()

        #res_data var used for while loop
        res_data = res.json()['data']

        complete_res_list.extend(res_data.get('items'))

        retrievedRows = res_data.get('totalRows')
        total_rows += retrievedRows

        counter = 0
        while 'nextPageToken' in res_data:
            #update payload
            json.update({'pageToken': res_data.get('nextPageToken')})

            #call page endpoint
            res = self.__api_call_wrapper('POST', f'{self.api_url}/search/config/page', json=json, verify=verify, acceptCsv=acceptCsv, redlock_ignore=redlock_ignore, status_ignore=status_ignore)
            counter += 1

            #update res_data with the paginated response
            res_data = res.json()

            #Add results from each page API call
            complete_res_list.extend(res_data.get('items'))

            retrievedRows = res_data.get('totalRows')
            total_rows+= retrievedRows

        #Update res dict to be the some format as the 'data' object in a typical RQL res
        complete_res_dict['data'].update({'totalRows':total_rows, 'items': complete_res_list})

        return complete_res_dict

    def config_search_request_function(self, json, function, verify=None, acceptCsv=False,  redlock_ignore: list=None, status_ignore: list=[]):
        if verify == None:
            verify = self.verify
        
        limit = 2000#Max limit value is 100,000
        
        #Force best practices with HS
        json.update({"heuristicSearch": True, "limit": limit, "withResourceJson": True})

        #initial API Call
        res = self.__api_call_wrapper('POST', f'{self.api_url}/search/config', json=json, verify=verify, acceptCsv=acceptCsv, redlock_ignore=redlock_ignore, status_ignore=status_ignore)
        
        total_rows = res.json()['data']['totalRows']

        res_details = res.json()
        res_details.pop('data')
        res_data = res.json()['data']
        counter = 0

        function(res_details, res_data, counter, total_rows)

        while 'nextPageToken' in res_data:
            #update payload
            json.update({'pageToken': res_data.get('nextPageToken')})

            #call page endpoint
            res = self.__api_call_wrapper('POST', f'{self.api_url}/search/config/page', json=json, verify=verify, acceptCsv=acceptCsv, redlock_ignore=redlock_ignore, status_ignore=status_ignore)
            counter += 1

            res_data = res.json()
            total_rows += res_data['totalRows']

            function(res_details, res_data, counter, total_rows)

        return total_rows

    def __request_wrapper(self, method, url, headers, json, data, params, verify, acceptCsv):
        if acceptCsv == True: #CSPM Support Only
            headers.update({
                'Accept': 'text/csv'
            })

        r = self.empty_res

        start_time = time.time()
        r = requests.request(method, url, headers=headers, json=json, data=data, params=params, verify=verify)
        end_time = time.time()
        time_completed = round(end_time-start_time,3)

        while r == self.empty_res and self.u_count < self.unknown_error_max:
            try:
                start_time = time.time()
                r = requests.request(method, url, headers=headers, json=json, data=data, params=params, verify=verify)
                end_time = time.time()
                time_completed = round(end_time-start_time,3)

                self.u_count = 1
                return [r, time_completed]
            except KeyboardInterrupt:
                self.logger.error('Keyboard Interrupt. Exiting...')
                exit()
            except Exception as e:
                self.logger.error(e)
                self.logger.error(f'UNKNOWN ERROR - Request Wrapper. Retrying {self.u_count} of {self.unknown_error_max}')
                time.sleep(2)
                self.u_count += 1

        return [r, time_completed]


#Installed
import requests

from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

#Local
from ._session_base import Session

import time

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class CSPMSession(Session):
    def __init__(self, tenant_name: str, a_key: str, s_key: str, api_url: str, verify, logger):
        """
        Initializes a Prisma Cloud API session for a given tenant.

        Keyword arguments:
        tenant_name -- Name of tenant associated with session
        a_key -- Tenant Access Key
        s_key -- Tenant Secret Key
        api_url -- API URL Tenant is hosted on
        """
        super().__init__(logger)

        self.tenant = tenant_name
        self.a_key = a_key
        self.s_key = s_key
        self.api_url = api_url
        self.verify = verify

        self.token_time_stamp = 0

        self.token = ''

        self.auth_key = 'x-redlock-auth'
        self.auth_style = ''

        self.headers = {
            'content-type': 'application/json; charset=UTF-8',
            'x-redlock-auth': ''
        }

        self._api_login_wrapper()

#==============================================================================     
    def _api_login(self) -> object:
        '''
        Calls the Prisma Cloud API to generate a x-redlock-auth JWT.
        Returns:
        x-redlock-auth JWT.
        '''

        #Build request
        url = f'{self.api_url}/login'
        
        headers = {
            'content-type': 'application/json; charset=UTF-8'
            }

        payload = {
            "username": f"{self.a_key}",
            "password": f"{self.s_key}"
        }

        self.logger.debug('API - Generating CSPM session token.')
        res = object()
        try:
            start_time = time.time()
            self.logger.debug(url)
            res = requests.request("POST", url, headers=headers, json=payload, verify=self.verify)
            end_time = time.time()
            time_completed = round(end_time-start_time,3)

            self.token_time_stamp = time.time()
        except:
            self.logger.error('Failed to connect to API.')
            self.logger.warning('Make sure any offending VPNs are disabled.')

        return [res, time_completed]

    def _expired_login(self) -> None:
        self.logger.error('FAILED')
        self.logger.warning('Invalid Login Credentials. JWT not generated. Exiting...')
        exit()

    def _api_refresh(self) -> None:
            self.logger.debug('API - Refreshing SaaS session token.')

            res = object()
            try:
                start_time = time.time()
                self.logger.debug(self.api_url + '/auth_token/extend')
                res = requests.request("GET", self.api_url + '/auth_token/extend', headers=self.headers, verify=self.verify)
                end_time = time.time()
                time_completed = round(end_time-start_time,3)
                self.token_time_stamp = time.time()
            except:
                self.logger.error('Failed to connect to API.')
                self.logger.warning('Make sure any offending VPNs are disabled.')

            return [res, time_completed]
